"""
Gradient checkpointing.
"""

from __future__ import annotations

from typing import List, Tuple
import typing
import contextlib
import weakref
import tensorflow as tf
from returnn.tf import compat as tf_compat
from returnn.tf.util import basic as tf_util


_grad_checkpoints = weakref.WeakKeyDictionary()  # type: weakref.WeakKeyDictionary[tf.Graph, List[Tuple[int, int]]]


@contextlib.contextmanager
def gradient_checkpoint_scope():
    """
    :return: context manager, where all tensors created inside the scope will be recomputed at backprop time,
        based on existing tensors which have been created earlier outside the scope.

    If prepare_gradient_checkpointing() is not called later, this does not have any effect.
    If no gradients are being calculated, this also does not have any effect.
    """
    graph = tf_compat.v1.get_default_graph()
    # noinspection PyProtectedMember
    first_op_id = graph._last_id + 1  # inclusive
    yield
    # noinspection PyProtectedMember
    last_op_id = graph._last_id + 1  # exclusive
    assert last_op_id > first_op_id
    for op_id in range(first_op_id, last_op_id):
        # noinspection PyProtectedMember
        op = graph._nodes_by_id[op_id]
        if getattr(op, "_RETURNN_gradient_checkpoint_exclude", False):
            continue
        op._RETURNN_gradient_checkpoint_first_op_id = first_op_id
    _grad_checkpoints.setdefault(graph, []).append((first_op_id, last_op_id))


@contextlib.contextmanager
def gradient_checkpoint_exclude_scope():
    """
    :return: context manager, where all tensors created inside the scope will be excluded
        for recomputation at backprop time.
    """
    graph = tf_compat.v1.get_default_graph()
    # noinspection PyProtectedMember
    first_op_id = graph._last_id + 1  # inclusive
    yield
    # noinspection PyProtectedMember
    last_op_id = graph._last_id + 1  # exclusive
    assert last_op_id >= first_op_id
    for op_id in range(first_op_id, last_op_id):
        # noinspection PyProtectedMember
        op = graph._nodes_by_id[op_id]
        if getattr(op, "_RETURNN_gradient_checkpoint_first_op_id", None) is not None:
            # There was some inner gradient_checkpoint_scope() again...
            continue
        op._RETURNN_gradient_checkpoint_exclude = True


def prepare_gradient_checkpointing():
    """
    Call this after the computation graph for calculating the model + loss has been created,
    before the gradients are calculated (before tf.gradients is called).

    This will create a copy of all the ops from within the gradient_checkpoint_scope() scope.

    This patches the op._gradient_function of all consuming ops
    to use the copied ops instead.
    So effectively, for backpropagation, it will recalculate all such tensors.
    """
    from tensorflow.python.framework import ops

    copied_ops = {}  # type: typing.Dict[tf.Operation, tf.Operation]  # old -> new

    # noinspection PyShadowingNames
    def _copy_op(op: tf.Operation) -> tf.Operation:
        if op in copied_ops:
            return copied_ops[op]

        new_inputs = []
        for x in op.inputs:
            x = _map_tensor(x)
            new_inputs.append(x)

        with tf_util.same_control_flow_ctx(op.outputs[0]), tf.name_scope(""):
            new_op = tf_util.copy_op(op, inputs=new_inputs, name=op.name)
        _set_wrapped_grad_func(new_op)
        copied_ops[op] = new_op
        return new_op

    def _map_tensor(x: tf.Tensor) -> tf.Tensor:
        assert isinstance(x, tf.Tensor)
        if getattr(x.op, "_RETURNN_gradient_checkpoint_first_op_id", None) is not None:
            x_op_copy = _copy_op(x.op)
            x = x_op_copy.outputs[x.value_index]
        return x

    # noinspection PyShadowingNames
    def _set_wrapped_grad_func(op: tf.Operation):
        # Make sure to not wrap it multiple times.
        if getattr(op, "_RETURNN_gradient_checkpoint_wrapped_grad_func", None) is not None:
            return

        try:
            orig_grad_func = ops.get_gradient_function(op)
        except LookupError:
            return  # no gradient function
        if orig_grad_func is None:
            return

        class _WrappedOp:
            def __init__(self):
                self.op = op
                self._inputs = tuple(_map_tensor(x) for x in op.inputs)
                self._outputs = tuple(_map_tensor(x) for x in op.outputs)

            @property
            def inputs(self) -> Tuple[tf.Tensor, ...]:
                """inputs"""
                return self._inputs

            @property
            def outputs(self) -> Tuple[tf.Tensor, ...]:
                """outputs"""
                return self._outputs

            def get_attr(self, name: str):
                """get_attr"""
                return self.op.get_attr(name)

        _wrapped_op = _WrappedOp()

        # noinspection PyShadowingNames
        def _wrapped_grad_func(op, *out_grads):
            assert op is _wrapped_op.op  # wrapped grad func not intended to be called for other ops
            return orig_grad_func(_wrapped_op, *out_grads)

        op._gradient_function = _wrapped_grad_func
        op._RETURNN_gradient_checkpoint_wrapped_grad_func = _wrapped_grad_func

    for graph, ls in _grad_checkpoints.items():
        for first_op_id, last_op_id in ls:
            for op_id in range(first_op_id, last_op_id):
                # noinspection PyProtectedMember
                op = graph._nodes_by_id[op_id]
                assert isinstance(op, tf.Operation)
                if getattr(op, "_RETURNN_gradient_checkpoint_exclude", False):
                    continue
                _copy_op(op)
                for op_in in op.inputs:
                    assert isinstance(op_in, tf.Tensor)
                    _set_wrapped_grad_func(op_in.op)
                for op_out in op.outputs:
                    assert isinstance(op_out, tf.Tensor)
                    for op_ in op_out.consumers():
                        assert isinstance(op_, tf.Operation)
                        _set_wrapped_grad_func(op_)



"""
Combine raw ops which are not directly supported by PyTorch.
"""

from __future__ import annotations

import torch


def squared_difference(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    """
    :param a:
    :param b:
    :return: (a - b) ** 2
    """
    return torch.square(a - b)



"""
Provides :class:`NumpyDumpDataset`.
"""

from returnn.datasets.basic import Dataset, DatasetSeq
import os
import numpy
import typing


class NumpyDumpDataset(Dataset):
    """
    For ``tools/dump-dataset.py --type=numpy``.
    """

    file_format_data = "%i.data"
    file_format_targets = "%i.targets"

    def __init__(
        self, prefix, postfix=".txt.gz", start_seq=0, end_seq=None, num_inputs=None, num_outputs=None, **kwargs
    ):
        super(NumpyDumpDataset, self).__init__(**kwargs)
        self.file_format_data = prefix + self.file_format_data + postfix
        self.file_format_targets = prefix + self.file_format_targets + postfix
        self.start_seq = start_seq
        self._init_num_seqs(end_seq)
        self._seq_index = None
        self.cached_seqs = []  # type: typing.List[DatasetSeq]
        self.num_inputs = num_inputs
        self.num_outputs = num_outputs
        assert num_inputs and num_outputs

    def _init_num_seqs(self, end_seq=None):
        last_seq = None
        i = self.start_seq
        while True:
            if end_seq is not None and i >= end_seq:
                break
            if not os.path.exists(self.file_format_data % i):
                break
            if not os.path.exists(self.file_format_targets % i):
                break
            last_seq = i
            i += 1
        if end_seq is None:
            assert last_seq is not None, "None found. Check %s." % (self.file_format_data % self.start_seq)
            end_seq = last_seq
        else:
            assert last_seq == end_seq - 1, "Check %s." % (self.file_format_data % end_seq)
        assert end_seq > self.start_seq
        self._num_seqs = end_seq - self.start_seq

    def _load_numpy_seq(self, seq_idx):
        """
        :param int seq_idx:
        """
        real_idx = self._seq_index[seq_idx]
        features = numpy.loadtxt(self.file_format_data % real_idx)
        targets = numpy.loadtxt(self.file_format_targets % real_idx)
        assert features.ndim == 2
        assert features.shape[1] == self.num_inputs
        assert targets.ndim == 1
        self._add_cache_seq(seq_idx, features, targets)

    # ------------ Dataset API --------------

    def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
        """
        :param int|None epoch:
        :param list[str]|None seq_list:
        :param list[int]|None seq_order:
        :rtype: bool
        """
        super(NumpyDumpDataset, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
        if seq_list is not None or seq_order is not None:
            raise NotImplementedError
        if self.seq_ordering == "sorted":  # not supported atm
            self.seq_ordering = "default"
        self._seq_index = [i + self.start_seq for i in self.get_seq_order_for_epoch(epoch, self.num_seqs)]
        self.cached_seqs[:] = []
        return True

    def _load_seqs(self, start, end):
        """
        :param int start:
        :param int end:
        """
        self._cleanup_old_seq_cache(start)
        for i in range(start, end):
            if not self._have_cache_seq(i):
                self._load_numpy_seq(i)

    def get_input_data(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: numpy.ndarray
        """
        return self._get_cache_seq(seq_idx).get_data("data")

    def get_targets(self, target, seq_idx):
        """
        :param str target:
        :param int seq_idx:
        :rtype: numpy.ndarray
        """
        return self._get_cache_seq(seq_idx).get_data(target)

    def get_seq_length(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: Util.NumbersDict
        """
        # This is different from the other get_* functions.
        # load_seqs() might not have been called before.
        if not self._have_cache_seq(seq_idx):
            self._load_numpy_seq(seq_idx)
        return self._get_cache_seq(seq_idx).num_frames

    @property
    def num_seqs(self):
        """
        :rtype: int
        """
        return self._num_seqs

    def len_info(self):
        """
        :rtype: str
        """
        return "%s, %i seqs" % (self.__class__.__name__, self.num_seqs)

    # ------------ Seq cache management -----------

    def _cleanup_old_seq_cache(self, seq_end):
        i = 0
        while i < len(self.cached_seqs):
            if self.cached_seqs[i].seq_idx >= seq_end:
                break
            i += 1
        del self.cached_seqs[:i]

    def _get_cache_seq(self, seq_idx, error_not_found=True):
        for data in self.cached_seqs:
            if data.seq_idx == seq_idx:
                return data
        if error_not_found:
            raise Exception("seq %i not loaded" % seq_idx)
        else:
            return None

    def _have_cache_seq(self, seq_idx):
        return self._get_cache_seq(seq_idx, error_not_found=False) is not None

    def _get_cache_last_seq_idx(self):
        if self.cached_seqs:
            return self.cached_seqs[-1].seq_idx
        else:
            return -1

    def _add_cache_seq(self, seq_idx, features, targets):
        last_seq_idx = self._get_cache_last_seq_idx()
        assert seq_idx == last_seq_idx + 1
        self.cached_seqs += [DatasetSeq(seq_idx, features, targets)]



#!/usr/bin/env python

"""
Collect orth symbols.
"""

from __future__ import annotations

import os
import sys

import _setup_returnn_env  # noqa
import returnn.__main__ as rnn
from returnn.log import log
from returnn.config import Config
import argparse
from returnn.util.basic import hms, human_size, parse_orthography, parse_orthography_into_symbols, unicode
import gzip
from xml.etree import ElementTree
import wave
import time


def found_sub_seq(sub_seq, seq):
    """
    :param list[str] sub_seq:
    :param list[str] seq:
    :rtype: bool
    """
    # Very inefficient naive implementation:
    for i in range(len(seq)):
        if seq[i : i + len(sub_seq)] == sub_seq:
            return True
    return False


def iter_dataset(dataset, options, callback):
    """
    :type dataset: Dataset.Dataset
    """
    dataset.init_seq_order(epoch=1)
    assert "orth" in dataset.get_target_list()

    seq_idx = 0
    while dataset.is_less_than_num_seqs(seq_idx):
        dataset.load_seqs(seq_idx, seq_idx)

        frame_len = dataset.get_seq_length(seq_idx)["data"]
        orth = dataset.get_targets("orth", seq_idx)
        callback(frame_len=frame_len, orth=orth)

        seq_idx += 1


def get_wav_time_len(filename):
    """
    :param str filename:
    :rtype: float
    """
    f = wave.open(filename)
    num_frames = f.getnframes()
    frame_rate = f.getframerate()
    f.close()
    return num_frames / float(frame_rate)


def iter_bliss(filename, options, callback):
    corpus_file = open(filename, "rb")
    if filename.endswith(".gz"):
        corpus_file = gzip.GzipFile(fileobj=corpus_file)

    def getelements(tag):
        """Yield *tag* elements from *filename_or_file* xml incrementaly."""
        context = iter(ElementTree.iterparse(corpus_file, events=("start", "end")))
        _, root = next(context)  # get root element
        tree = [root]
        for event, elem in context:
            if event == "start":
                tree += [elem]
            elif event == "end":
                assert tree[-1] is elem
                tree = tree[:-1]
            if event == "end" and elem.tag == tag:
                yield tree, elem
                root.clear()  # free memory

    time_via_wav = False

    for tree, elem in getelements("segment"):
        if options.collect_time:
            start = float(elem.attrib.get("start", 0))
            if "end" in elem.attrib:
                end = float(elem.attrib["end"])
            else:
                if not time_via_wav:
                    time_via_wav = True
                    print(
                        "Time will be read from WAV recordings. Can be slow. Maybe use `--collect_time False`.",
                        file=log.v3,
                    )
                rec_elem = tree[-1]
                assert rec_elem.tag == "recording"
                wav_filename = rec_elem.attrib["audio"]
                end = get_wav_time_len(wav_filename)
            assert end > start
            frame_len = (end - start) * (1000.0 / options.frame_time)
        else:
            frame_len = 0
        elem_orth = elem.find("orth")
        orth_raw = elem_orth.text or ""  # should be unicode
        orth_split = orth_raw.split()
        orth = " ".join(orth_split)

        callback(frame_len=frame_len, orth=orth)


def iter_txt(filename, options, callback):
    f = open(filename, "rb")
    if filename.endswith(".gz"):
        f = gzip.GzipFile(fileobj=f)

    if options.collect_time:
        print("No time-info in txt.", file=log.v3)
        options.collect_time = False

    for line in f:
        line = line.strip()
        if not line:
            continue

        callback(frame_len=0, orth=line)


def collect_stats(options, iter_corpus):
    """
    :param options: argparse.Namespace
    """
    orth_symbols_filename = options.output
    if orth_symbols_filename:
        assert not os.path.exists(orth_symbols_filename)

    class Stats:
        count = 0
        process_last_time = time.time()
        total_frame_len = 0
        total_orth_len = 0
        orth_syms_set = set()

    if options.add_numbers:
        Stats.orth_syms_set.update(map(chr, list(range(ord("0"), ord("9") + 1))))
    if options.add_lower_alphabet:
        Stats.orth_syms_set.update(map(chr, list(range(ord("a"), ord("z") + 1))))
    if options.add_upper_alphabet:
        Stats.orth_syms_set.update(map(chr, list(range(ord("A"), ord("Z") + 1))))

    def cb(frame_len, orth):
        if frame_len >= options.max_seq_frame_len:
            return
        orth_syms = parse_orthography(orth)
        if len(orth_syms) >= options.max_seq_orth_len:
            return

        Stats.count += 1
        Stats.total_frame_len += frame_len

        if options.dump_orth_syms:
            print("Orth:", "".join(orth_syms), file=log.v3)
        if options.filter_orth_sym:
            if options.filter_orth_sym in orth_syms:
                print("Found orth:", "".join(orth_syms), file=log.v3)
        if options.filter_orth_syms_seq:
            filter_seq = parse_orthography_into_symbols(options.filter_orth_syms_seq)
            if found_sub_seq(filter_seq, orth_syms):
                print("Found orth:", "".join(orth_syms), file=log.v3)
        Stats.orth_syms_set.update(orth_syms)
        Stats.total_orth_len += len(orth_syms)

        # Show some progress if it takes long.
        if time.time() - Stats.process_last_time > 2:
            Stats.process_last_time = time.time()
            if options.collect_time:
                print(
                    "Collect process, total frame len so far:",
                    hms(Stats.total_frame_len * (options.frame_time / 1000.0)),
                    file=log.v3,
                )
            else:
                print("Collect process, total orth len so far:", human_size(Stats.total_orth_len), file=log.v3)

    iter_corpus(cb)

    if options.remove_symbols:
        filter_syms = parse_orthography_into_symbols(options.remove_symbols)
        Stats.orth_syms_set -= set(filter_syms)

    if options.collect_time:
        print(
            "Total frame len:",
            Stats.total_frame_len,
            "time:",
            hms(Stats.total_frame_len * (options.frame_time / 1000.0)),
            file=log.v3,
        )
    else:
        print("No time stats (--collect_time False).", file=log.v3)
    print("Total orth len:", Stats.total_orth_len, "(%s)" % human_size(Stats.total_orth_len), end=" ", file=log.v3)
    if options.collect_time:
        print("fraction:", float(Stats.total_orth_len) / Stats.total_frame_len, file=log.v3)
    else:
        print("", file=log.v3)
    print("Average orth len:", float(Stats.total_orth_len) / Stats.count, file=log.v3)
    print("Num symbols:", len(Stats.orth_syms_set), file=log.v3)

    if orth_symbols_filename:
        orth_syms_file = open(orth_symbols_filename, "wb")
        for orth_sym in sorted(Stats.orth_syms_set):
            orth_syms_file.write(b"%s\n" % unicode(orth_sym).encode("utf8"))
        orth_syms_file.close()
        print("Wrote orthography symbols to", orth_symbols_filename, file=log.v3)
    else:
        print("Provide --output to save the symbols.", file=log.v3)


def init(config_filename=None):
    rnn.init_better_exchook()
    rnn.init_thread_join_hack()
    if config_filename:
        rnn.init_config(config_filename, command_line_options=[])
        rnn.init_log()
    else:
        log.initialize()
    print("RETURNN collect-orth-symbols starting up.", file=log.v3)
    rnn.init_faulthandler()
    if config_filename:
        rnn.init_data()
        rnn.print_task_properties()


def is_bliss(filename):
    try:
        corpus_file = open(filename, "rb")
        if filename.endswith(".gz"):
            corpus_file = gzip.GzipFile(fileobj=corpus_file)
        context = iter(ElementTree.iterparse(corpus_file, events=("start", "end")))
        _, root = next(context)  # get root element
        return True
    except IOError:  # 'Not a gzipped file' or so
        pass
    except ElementTree.ParseError:  # 'syntax error' or so
        pass
    return False


def is_crnn_config(filename):
    if filename.endswith(".gz"):
        return False
    try:
        config = Config()
        config.load_file(filename)
        return True
    except Exception:
        pass
    return False


def main(argv):
    argparser = argparse.ArgumentParser(description="Collect orth symbols.")
    argparser.add_argument("input", help="RETURNN config, Corpus Bliss XML or just txt-data")
    argparser.add_argument(
        "--frame_time", type=int, default=10, help="time (in ms) per frame. not needed for Corpus Bliss XML"
    )
    argparser.add_argument(
        "--collect_time", type=int, default=True, help="collect time info. can be slow in some cases"
    )
    argparser.add_argument("--dump_orth_syms", action="store_true", help="dump all orthographies")
    argparser.add_argument("--filter_orth_sym", help="dump orthographies which match this filter")
    argparser.add_argument("--filter_orth_syms_seq", help="dump orthographies which match this filter")
    argparser.add_argument(
        "--max_seq_frame_len", type=int, default=float("inf"), help="collect only orthographies <= this max frame len"
    )
    argparser.add_argument(
        "--max_seq_orth_len", type=int, default=float("inf"), help="collect only orthographies <= this max orth len"
    )
    argparser.add_argument("--add_numbers", type=int, default=True, help="add chars 0-9 to orth symbols")
    argparser.add_argument("--add_lower_alphabet", type=int, default=True, help="add chars a-z to orth symbols")
    argparser.add_argument("--add_upper_alphabet", type=int, default=True, help="add chars A-Z to orth symbols")
    argparser.add_argument("--remove_symbols", default="(){}$", help="remove these chars from orth symbols")
    argparser.add_argument("--output", help="where to store the symbols (default: dont store)")
    args = argparser.parse_args(argv[1:])

    bliss_filename = None
    crnn_config_filename = None
    txt_filename = None
    if is_bliss(args.input):
        bliss_filename = args.input
    elif is_crnn_config(args.input):
        crnn_config_filename = args.input
    else:  # treat just as txt
        txt_filename = args.input
    init(config_filename=crnn_config_filename)

    if bliss_filename:
        iter_corpus = lambda cb: iter_bliss(bliss_filename, options=args, callback=cb)
    elif txt_filename:
        iter_corpus = lambda cb: iter_txt(txt_filename, options=args, callback=cb)
    else:
        iter_corpus = lambda cb: iter_dataset(rnn.train_data, options=args, callback=cb)
    collect_stats(args, iter_corpus)

    if crnn_config_filename:
        rnn.finalize()


if __name__ == "__main__":
    main(sys.argv)
